<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
    <title>Evaluation Report - UVal - {{ title }}</title>
    <style type="text/css">
        .reportBody{text-align:center;}
        * {font-family: sans-serif;}
        body {padding: 20px;}
        table {border-collapse: collapse;text-align: right;width: 100%;}
        table tr {border-bottom: 1px solid}
        table th, table td {padding: 10px 20px;}
        .custom_li {list-style-type:none;border-bottom: 1px solid gray;}
        .header, .footer {min-width:100%;line-height:2;text-align:center;border-radius:5px;}
        .header {background-color:#0065cb;color: white}
        .footer {background-color:#8bbae8}
    </style>
</head>
    <body>
        <div class="reportBody">
            <h1 class="header">Evaluation Report {{ title }}</h1>
            {{ single_table }}
            <br/>
            <ul>
                {% for img_names in total_images %}
                    <li class="custom_li" >
                        {% for img_name in img_names %}
                            <img src={{img_name}}>
                        {% endfor %}
                    </li>
                {% endfor %}
            </ul>
            <br />
        </div>
    </body>
    <h5 class="footer">Powered by <a href="https://uval.github.io/">UVal</a></h5>
</html>
